//******************************************************************************************
//* Mandelbrot and Julia Set Fractal Program, by Paul Nylander, bugman123.com, 6/8/02?
//******************************************************************************************

// In order to compile with OpenGL, you will need these files: http://www.ics.uci.edu/~frost/past/ics186a/EssentialFiles.zip
// Simply place the files in these folders: http://www.ics.uci.edu/~frost/past/ics186a/wheretheygo.html

#include <string.h> // strlen
#include <GL/glut.h>
#include <stdio.h>
#include <math.h>

// menu constants
#define MANDELBROT 0 // set
#define JULIA      1
#define ITERATIONS 2 // method
#define MAGNITUDE  3
#define REAL       4
#define IMAGINARY  5
#define RED        6 // color
#define YELLOW     7
#define GREEN      8
#define CYAN       9
#define BLUE       10
#define MAGENTA    11
#define WHITE      12
#define QUIT       13

int set = MANDELBROT;
int nx, ny, imax=50, method=ITERATIONS, color=RED, hue=0;
double x0=-0.5, y0a=0, dpi, power=2, xc=0.5, yc=0.05;

double abs2(double x) {return (x>0)?x:(-x);}
double min(double x1, double x2) {return (x1<x2)?x1:x2;}
double max(double x1, double x2) {return (x1>x2)?x1:x2;}
double rem(double x1, double x2) {return x1-x2*(int)(x1/x2);}
// double sinh(double x) {double t1 = exp(x); return (t1 - 1.0/t1) / 2.0;}
// double cosh(double x) {double t1 = exp(x); return (t1 + 1.0/t1) / 2.0;}

void pixel(int x, int y, int r, int g, int b) {
 glColor3f(r/256.0, g/256.0, b/256.0);
 glVertex2i(x, y);
}

void text(int x, int y, char *txt) {
 int n=(int)strlen(txt), i;
 glRasterPos2f(x, y); // cursor location
 for (i=0; i<n; i++) {glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24, txt[i]);}
}

void display(void) {
 int unbounded = 0;
 int r=0, g=0, b=0, i, n=(int)power/2-((rem(power,2)!=0)?1:0);
 double x=0, y=0, xc1=xc, yc1=yc, dx=1/dpi, x1=x0-0.5*nx/dpi, x2=x0+0.5*nx/dpi, y1a=y0a-ny*dx/2, y2=y0a+ny*dx/2;
 double t1=max(abs2(x1),abs2(x2)), t2=max(abs2(y1a),abs2(y2)), zm=sqrt(t1*t1+t2*t2);
 int iy,ix,j; // supposed to go in for loop
 double r1, g1, b1;
 char txt[100];
 switch(color) { // calculations
  case RED:     hue=0  ; break;
  case YELLOW:  hue=60 ; break;
  case GREEN:   hue=120; break;
  case CYAN:    hue=180; break;
  case BLUE:    hue=240; break;
  case MAGENTA: hue=300; break;
 }
 r1=max(0,min(1,abs2(rem(6-hue/60.0,6)-3)-1));
 g1=max(0,min(1,abs2(rem(8-hue/60.0,6)-3)-1));
 b1=max(0,min(1,abs2(rem(10-hue/60.0,6)-3)-1));
 glClear(GL_COLOR_BUFFER_BIT);
 glBegin(GL_POINTS);

 for (iy=0; iy<=ny; ++iy) {
  if (set==MANDELBROT) yc1=y2-iy*dx;
  for (ix=0; ix<=nx; ++ix) {
   if (set==JULIA) {x=x1+ix*dx; y=y2-iy*dx;} else {xc1=x1+ix*dx; x=0; y=0;}
   unbounded=0;
   for (i=0; (i<imax)&&(!unbounded); ++i) {
    if (power>=2 && power<=25 && rem(power,1)==0) {
     for (j=0; j<n; ++j) {t1=x*x-y*y; y=2*x*y; x=t1;}; // z^2
     if (rem(power,2)!=0) {t1=x*(x*x-3*y*y); y=y*(3*x*x-y*y); x=t1;} // z^3
    }
    else {t1= atan2(y,x); t2= pow(x*x+y*y,power/2.0); x=t2* cos(power*t1); y=t2* sin(power*t1);} // z^n
//  t1=x*x+y*y; x=x/t1; y=-y/t1; // z^-1
//  t1=atan(y/x); t2= pow(x*x+y*y,0.25); x=t2* cos(t1/2); y=t2* sin(t1/2); // sqrt(z)
//  t1=sin(x)*cosh(y); y=cos(x)*sinh(y); x=t1; // sin(z)
//  t1=cos(x)*cosh(y); y=-sin(x)*sinh(y); t1=x; // cos(z)
//  t1=cos(2*x)+cosh(2*y); x=sin(2*x)/t1; y=sinh(2*y)/t1; // tan(z)
//  t1=exp(x); x=t1* cos(y); y=t1*sin(y); // e^z
//  t1=log(x*x+y*y)/2; y= atan(y/x); x=t1; // ln(z)
    x+=xc1; y+=yc1;
    if (x*x+y*y>=4.0) unbounded=1;
   }
   if (unbounded) {
    if (method==ITERATIONS) t1=1.0*i/imax;
    else if (method==MAGNITUDE) t1= min(1, sqrt(x*x+y*y)/zm);
    else t1= min(1, abs2((method==REAL)?x:y)/zm);
    if (color==WHITE) {r=255*t1; g=r; b=r;}
    else { // Note: brightness=(int)t1*100;
     r = 255*((t1>0.5)?t1*r1+(1-t1*r1)*(2*t1-1):2*t1*r1);
     g = 255*((t1>0.5)?t1*g1+(1-t1*g1)*(2*t1-1):2*t1*g1); 
     b = 255*((t1>0.5)?t1*b1+(1-t1*b1)*(2*t1-1):2*t1*b1); 
    }
    pixel(ix,iy,r,g,b);
   }
  }
 }

 glEnd();
 glColor3f(1,1,1);
 sprintf(txt, "%s Set: %f", (set == JULIA) ? "Julia" : "Mandelbrot", t1);

 text(10, 10, txt);
 glutSwapBuffers();
}

void mousemove(int x, int y) {
 glutPostRedisplay();
}

// Menu
void HandleMenu(int menuChoice) {
 switch (menuChoice) {
 case MANDELBROT: case JULIA: set = menuChoice; power=(set==JULIA)?3:2; break;
  case ITERATIONS: case MAGNITUDE: case REAL: case IMAGINARY: method = menuChoice; break;
  case RED: case YELLOW: case GREEN: case CYAN: case BLUE: case MAGENTA: case WHITE: color = menuChoice; break;
  case QUIT: exit(0);
 }
 glutPostRedisplay();
}

int main(int argc, char** argv) {
 int setSubMenu, methodSubMenu, colorSubMenu;
 nx=640; ny=nx*3/4;
 dpi=nx/3.0;
 glutInit(&argc, argv);
 glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
 glutInitWindowSize(nx, ny);
 glutInitWindowPosition(25, 50);
 glutCreateWindow("Mandelbrot");
 gluOrtho2D(0, nx, 0, ny);
 glutDisplayFunc(display);
 glClearColor(0, 0, 0, 1);

 setSubMenu = glutCreateMenu(HandleMenu); // set sub-menu
 glutAddMenuEntry("Mandelbrot Set", MANDELBROT);
 glutAddMenuEntry("Julia Set", JULIA);

 methodSubMenu = glutCreateMenu(HandleMenu); // method sub-menu
 glutAddMenuEntry("Iterations", ITERATIONS);
 glutAddMenuEntry("Magnitude", MAGNITUDE);
 glutAddMenuEntry("Real", REAL);
 glutAddMenuEntry("Imaginary", IMAGINARY);

 colorSubMenu = glutCreateMenu(HandleMenu); // color sub-menu
 glutAddMenuEntry("Red", RED);
 glutAddMenuEntry("Green", GREEN);
 glutAddMenuEntry("Blue", BLUE);
 glutAddMenuEntry("Yellow", YELLOW);
 glutAddMenuEntry("Cyan", CYAN);
 glutAddMenuEntry("Magenta", MAGENTA);
 glutAddMenuEntry("White", WHITE);

 // menu
 glutCreateMenu(HandleMenu);
 glutAddSubMenu("Set", setSubMenu);
 glutAddSubMenu("Method", methodSubMenu);
 glutAddSubMenu("Color", colorSubMenu);
 glutAddMenuEntry("Quit", QUIT);
 glutAttachMenu(GLUT_RIGHT_BUTTON);

 glutMainLoop();
 return 0;
}
